/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

package com.funambol.mail;

import com.funambol.util.StringUtil;

/**
 * A class describing the RFC 2822 address specification as per  3.4
 */
public class Address {

    // -------------------------------------------------------------- Constants
 
    public static final int FROM = 1;

    public static final int TO = 2;

    public static final int CC = 3;

    public static final int BCC = 4;

    public static final int REPLYTO = 5;

    // ----------------------------------------------------------- Private Data

    /**
     * The type of the address
     */
    private int type;

    /**
     * The optional display name (part 1 of the mailbox)
     */
    private String name;
    
    /**
     * The email address (part 2 of the mailbox)
     */
    private String email;

    // ------------------------------------------------------------- Properties

    // ----------------------------------------------------------- Constructors

    /**
     * Creates a new instance of Address, parsing <code>address<code/>
     * to
     */
    public Address(int type, String address) throws MailException {
        this.type = type;
        parseAddr(address);
    }


    /**
     * Creates a new instance of Address
     */
    public Address(int type, String name, String address) throws MailException {
        this.type = type;
        this.name = name;
        this.email= checkAddr(address);
    }


    // --------------------------------------------------------- Public Methods

    /**
     * Returns the type of address
     */
    public int getType() {
        return type;
    }

    /**
     * Returns the name or the address if name is empy
     */ 
    public String getVisibleName() {
        return name != null ? name : email;
    }
    
    /**
     * Returns the visible name TODO: or the address if null?
     */
    public String getName() {
        return name != null ? name : "";
    }


    /**
     * Returns the email address
     */
    public String getEmail() {
        return email;
    }


    /**
     * Set the type of address
     */
    public void setType(int type) {
        this.type = type;
    }


    /**
     * Set the visible name
     */
    public void setName(String name) {
        this.name = name;
    }


    /**
     * Set the email address
     */
    public void setEmail(String email) throws MailException {
        this.email = checkAddr(email);
    }


    /**
     * Parses an address list and return an array of <code>Address</code>
     * objects, each representing a recipient.
     * 
     * Use: Address [] list = Address.parse( Address.TO, "John
     * Doe&lt;john.doe@mail.com&gt;"
     * 
     * @param addresslist
     *            the comma separated list of addresses
     * @return an array of <code>Address</code> containing the result, or an
     * empty array if addresslist is empty
     * 
     * @throws MailException
     *             if the address is malformed
     *
     * TODO: Improve the Parser: now a visible name like:
     *          "Mario Rossi, Funambol Inc" breaks it.
     */
    public static Address[] parse(int type, String addresslist)
            throws MailException {
      
        if (addresslist == null || addresslist.trim().length() == 0)
            return new Address[0];
        
        String[] addrlist = StringUtil.split(addresslist, ",");
        
        int len = addrlist.length;
        Address[] ret = new Address[len];

        for (int i = 0; i < len; i++) {
            String addr = addrlist[i].trim();
            if (addr != null && !addr.equals("")) {
                ret[i] = new Address(type, addr);
            }
        }
        return ret;
    }


    /**
     * Returns the string representation of the Address, in RFC2822 format.
     */
    public String toString() {
        // If a name is defined, compose the '"name" <address>' form
        if (name != null) {
            StringBuffer ret = new StringBuffer("");

            ret.append("\"").append(name).append("\"").
                append(" <").append(email).append(">");

            return ret.toString();
        }

        // Otherwise return the address
        return email;
    }


    // -------------------------------------------------------- Private Methods

    // Parse a single address
    private void parseAddr(String address) throws MailException {

        if (address == null) {
            throw new MailException("Invalid address (null)");
        }
        
        int quote = address.indexOf("\"");
        int op_bracket = address.indexOf("<");
        int cl_bracket = address.indexOf(">");        
        if (op_bracket != -1) {
            if (cl_bracket == -1) {
                throw new MailException("Invalid address: " + address);
            }
            this.email = address.substring(op_bracket + 1, cl_bracket);
            if (quote != -1) {
                int cl_quote = address.indexOf("\"", quote + 1);
                if (cl_quote == -1) {
                    throw new MailException("Invalid address: " + address);
                }
                this.name = address.substring(quote + 1, cl_quote);

            } else {
                // The first character isn't '"'
                // (e.g. Pinco Pallino <pinco@pallino.com>)
                if (op_bracket != 0) { // '<' isn't the first character
                    // Pinco Pallino
                    this.name = address.substring(0, op_bracket).trim();
                }
            }

        } else {
            this.name = null;
            this.email = address;
        }
        
        // Perform a simple validity check on the address.
        checkAddr(address);
    }

    /**
     * check if the address is valid
     */
    private String checkAddr(String address) throws MailException {
        int position=address.indexOf("@");
       
        if ( (position!=-1) && (position!=0) && (position!=address.length())) {
            return address;

        } else {
            throw new MailException("checkAddr: Invalid email address: "
                                    + address);
        }
    }
}

