/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.mail;

import java.util.Date;

import com.funambol.util.Base64;
import com.funambol.util.Log;
import com.funambol.util.MailDateFormatter;
import com.funambol.util.StringUtil;

/**
 * Provides the functionality to transform a <code>Message</code> object into
 * a string formatted as per RFC 2822/MIME ready to be passed to the transport
 * layer of the email application
 */
public class MIMEFormatter {

    // -------------------------------------------------------------- Constants

    private static final String CC = "Cc: ";
    
    private static final String CONTENT_DISPOSITION = "Content-Disposition: ";
    
    private static final String CONTENT_TYPE = "Content-Type: ";

    private static final String CONTENT_TRANSFER_ENCODING =
        "Content-Transfer-Encoding: ";
    
    private static final String BCC = "Bcc: ";

    private static final String DATE = "Date: ";

    private static final String FILENAME = "filename=";
    
    private static final String FROM = "From: ";
    
    private static final String MESSAGE_ID = "Message-ID: ";
    /** The MIME version implemented by this processor */
    private static final String MIME_VERSION = "MIME-Version: 1.0";
    
    private static final String REPLY_TO = "Reply-To: ";

    private static final String SUBJECT = "Subject: ";

    private static final String TO = "To: ";

    private static final String UA = "User-Agent: ";

    private static final String CRLF = "\r\n";

    // --------------------------------------------------------- Private Fields

    /**
     * The boundary of a multi-part <code>Message</code>. Initialized to an
     * empty string to be used by building single-part Messages too
     */
    private String boundary = "";

    /**
     * "--boundary"
     */
    private String startboundary = "";

    /**
     * "--boundary--"
     */
    private String endboundary = "";

    /**
     * Holds a reference to the <code>Message</code> that has to be formatted
     * into a RFC 2822/MIME string
     */
    private Message mailmessage = null;

    /**
     * The content of the "To:" header attribute
     */
    private String recipients = null;

    /**
     * The textual content of a 'single-part' <code>Message</code>
     */
    private String textContent = null;

    /**
     * The content of a <code>BodyPart</code>
     */
    private String bodyPartContent = null;

    /**
     * The <code>BodyPart</code>s of this multi-part <code>Message</code>
     */
    private StringBuffer partString;

    /**
     * Just an integer to increase the randomness of the boundary string created
     * with #createUniqueBoundaryValue()
     */
    private int part = 0;
    

    // ----------------------------------------------------------- Constructors

    /**
     * Initializes the fields containing a reference to the <code>Message</code>
     * object and to the name of the <code>Folder</code> containing this
     * <code>Message</code> object
     * 
     * @param mailmessage
     *            The <code>Message</code> object that has to be formatted
     *            into a RFC 2822/MIME string. This is necessary in order to
     *            allow a consumer to use the <code>toString()</code> method
     *            of this formatter after having instantiated it
     * 
     * TODO: check if we really need the folder name
     */
    public MIMEFormatter(Message mailmessage) {
        this.mailmessage = mailmessage;
    }

    // --------------------------------------------------------- Public Methods

    /**
     * The values of the fields of the <code>Message</code> object referenced
     * by 'mailmessage' are analyzed to be estracted and formatted into a string
     * in the form foreseen by the RFC 2822/MIME specifications
     * 
     */
    public String toString() {
        // message ID created in Message()
        String messageID = mailmessage.getMessageId();

        // the subject
        String subject = mailmessage.getSubject();
        if (subject == null)
            subject = "";// we explicitly don't want any "(no subject)" warn

        // TODO: this is valid only if the originator has only one address
        String from = mailmessage.getHeader(Message.FROM);

        // origination date field: see RFC 2822,  3.6.1
        Date date = mailmessage.getSentDate();

        // User-Agent
        String useragent = mailmessage.getHeader(UA);
        //useragent = useragent != null ? useragent : "";

        /*
         * TODO: handle the content-transfer-encoding contained in the body part
         * of the message
         */
        String contentTransferEncoding =
            mailmessage.getHeader(Message.CONTENT_TRANSFER_ENCODING);
        contentTransferEncoding =
            contentTransferEncoding != null ? contentTransferEncoding : "8bit";

        /*
         * this is only [meta type]/[subtype], e.g. "multipart/mixed" or
         * "text/plain"
         */
        String contentType = mailmessage.getContentType() + "; charset=UTF-8";

        // handling multipart messages
        if (mailmessage.getContent() instanceof Multipart) {

            if (mailmessage.isMultipart()) {
                boundary = createUniqueBoundaryValue();
                startboundary = "--" + boundary + CRLF;
                endboundary = "--" + boundary + "--";
            }

            Multipart multipart = (Multipart)mailmessage.getContent();
            partString = new StringBuffer();

            int size = multipart.getCount();
            for (int i = 0; i < size; i++) {
                BodyPart bodypart = multipart.getBodyPart(i);

                // TODO: handle recursive multiparts (use 'isMultipart()')
                
                /*
                 * TODO: this is the original content type, because it is added
                 * by the MIMEParser to the headers of the body part. But in
                 * BodyPart the method setContent provides default media types
                 * (like "application/octet-stream"), that are stored in the
                 * 'contentType' field but not used here. The implementation of
                 * setContent in BodyPart has to be changed
                 */
                String ct = bodypart.getContentType();
                String cte = bodypart.getHeader(BodyPart.CONTENT_TRANSFER_ENCODING);
                String cd = bodypart.getDisposition();
                String fn = bodypart.getFileName();
                
                // adding headers to part (CONTENT_TYPE is "Content-Type: ")
                partString.append(CONTENT_TYPE + ct + CRLF).append(
                    CONTENT_TRANSFER_ENCODING
                            + (!cte.equals("base64") ? "utf-8" : cte) + CRLF)
                        .append(CONTENT_DISPOSITION + cd);
                if (fn != null) {
                    partString.append("; " + FILENAME + fn + CRLF + CRLF);
                } else {
                    partString.append(CRLF + CRLF);
                }        

                // adding content to part
                if (bodypart.getContent() instanceof String) {
                    bodyPartContent = (String)bodypart.getContent();
                } else if (bodypart.getContent() instanceof byte[]
                        && cte.equals("base64")) {
                    bodyPartContent =
                        new String(Base64.encode((byte[])bodypart.getContent()));
                    //each line of the base64 block hasn't to exceed 76 chars
                    int len = bodyPartContent.length();
                    if (len > 76) {
                        StringBuffer bpc = new StringBuffer();
                        for (int j = 0, k = j + 76; j < len; j += 76 + 2, k = j + 76) {//+2 beacuse of '\r' and '\n'
                            if (len - j < 76)
                                k = len;
                            bpc.append(bodyPartContent.substring(j, k) + CRLF);
                        }
                        bodyPartContent = bpc.toString();
                    }
                } else {// TODO: handle recursive multiparts
                    bodyPartContent =
                        "Warning! The content of this part isn't of a known type!";
                    Log.error(bodyPartContent);
                }

                // TODO: check if we have to add a blank (probably not)
                partString.append(bodyPartContent);
                if (i != size - 1) {
                    partString.append("--" + boundary + CRLF);
                }
            }
        } else {
            // TODO: handle multipart content too
            textContent = mailmessage.getTextContent();
            if (textContent == null) {
                textContent =
                    "Error: the content of the message isn't text!";
                Log.error("The content of the message isn't text!");
            }
        }

        StringBuffer sb = new StringBuffer();
        
        String content =
            partString != null ? partString.toString() : textContent;
        
        sb.append(DATE + MailDateFormatter.dateToRfc2822(date) + CRLF).
            append(FROM + from + CRLF);

        if(useragent != null) {
            sb.append(UA + useragent + CRLF);
        }

        // Append MIME-Version implemented by this MIME Processor.
        sb.append(MIME_VERSION + CRLF);
            
        String to = mailmessage.getHeader(Message.TO);
        if (to != null)
            sb.append(TO + StringUtil.fold(to));
        
        String cc = mailmessage.getHeader(Message.CC);
        if (cc != null && !"".equals(cc))
            sb.append(CC + StringUtil.fold(cc));

        String replyto = mailmessage.getHeader(Message.REPLYTO);
        if (replyto != null && !"".equals(replyto))
            sb.append(REPLY_TO + StringUtil.fold(replyto));
        
        String bcc = mailmessage.getHeader(Message.BCC);
        if (bcc != null && !"".equals(bcc))
            sb.append(BCC + StringUtil.fold(bcc));
        
        sb.append(SUBJECT + subject + CRLF).
            append(MESSAGE_ID + messageID + CRLF).
            append(CONTENT_TYPE + contentType + CRLF).
            append(CONTENT_TRANSFER_ENCODING + contentTransferEncoding + CRLF).
            append(CRLF).
            append(startboundary).
            append(content).append(CRLF).
            append(endboundary);

        //Log.debug("partString is: " + partString);
        //Log.debug(sb.toString());

        return sb.toString();
    }


    // -------------------------------------------------------- Private Methods

    /**
     * Creates a unique boundary as described in the RFC 2046
     * 
     * @return A unique boundary as described in the RFC 2046
     */
    private String createUniqueBoundaryValue() {
        StringBuffer boundary = new StringBuffer();
        boundary.append("--=_Part_").append(part++).append("_").append(
            boundary.hashCode()).append('.').append(System.currentTimeMillis());
        return boundary.toString();
    }

}
