/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.mail;

import java.util.Vector;
import com.funambol.storage.NamedObjectStore;
import com.funambol.storage.Serializable;

import com.funambol.mail.LightMessage;
import com.funambol.util.Log;

import java.io.IOException;
import javax.microedition.rms.RecordStore;
import javax.microedition.rms.InvalidRecordIDException;
import javax.microedition.rms.RecordStoreException;
import javax.microedition.rms.RecordStoreNotFoundException;

/**
 * Implements the Store using J2ME RecordStore. <p> A folder is mapped to a
 * RecordStore with name equal to the full path of the folder, and the messages
 * inside the folder are the records inside that RecordStore. <p> It uses
 * NamedObjectStore to speed up the record access by messageId.
 */
public class RMSStore implements Store {

    // -------------------------------------------------------------- Attributes
    private NamedObjectStore objs;

    private Folder folder;

    private Message message;

    // Automatic storage handling:
    private int lo_watermark = 16*1024; // Leave of RMS 16K free
    private int hi_watermark = 32*1024; // and cleanup up to 32.

    // ------------------------------------------------------------ Constructors
    RMSStore() {
        objs = new NamedObjectStore();
        folder = null;
        message = null;
    }


    // ---------------------------------------------------------- Public methods

    /**
     * Initialize the message store, creating the main folders.
     *
     * @param reset if true, erase and re-create all the main folders.
     */
    public void init(boolean reset) throws StoreException {
        if (reset) {
            removeFolder(INBOX);
            removeFolder(OUTBOX);
            removeFolder(DRAFTS);
            removeFolder(SENT);
            removeFolder(TRASH);
        }

        createFolder(INBOX);
        createFolder(OUTBOX);
        createFolder(DRAFTS);
        createFolder(SENT);
        createFolder(TRASH);
    }

    /**
     * This method returns the list of the top level folders in this store.
     * 
     * @return an array of folders, or <code>null</code> if the store is
     *         empty.
     */
    public Folder[] list() {
        try {
            return list("//");
        } catch (StoreException e) {
            e.printStackTrace(); // Cannot happen
        }
        return null;
    }


    /**
     * This method returns the list of the folders whose path starts with 'path'
     * and are direct subfolders of it.
     * 
     * @param path
     *            the path of the parent folder
     * @return an array of folders, or <code>null</code> if <code>path</code>
     *         has no subfolders.
     * 
     * @throws StoreException
     *             If the path is not valid
     */
    public Folder[] list(String path) throws StoreException {
        path = checkPath(path);
        String[] folders = RecordStore.listRecordStores();
        Vector ret = new Vector();

        for (int i = 0; i < folders.length; i++) {
            if (folders[i].startsWith(path)
                    && folders[i].indexOf('/', path.length()) == -1) {
                ret.addElement(new Folder(folders[i], this));
            }
        }
        return toFolderArray(ret);
    }


    /**
     * This method creates a folder in the record store. The names of the
     * folder must start with a "/" to distinguish them (they are record stores)
     * from other record stores. The method add the leading slash if not 
     * present.
     * 
     * @param path
     *             the full path of the folder in the Store.
     * @throws StoreException
     *             If an error occurs on the store (e.g. no space left)
     */
    public Folder createFolder(String path) throws StoreException {
        path = checkPath(path);
        try {
            objs.create(path);
            return new Folder(path, this);
        } catch (RecordStoreException rse) {
            rse.printStackTrace();
            throw new StoreException("[createFolder] Can't create folder: "
                    + path);
        } catch (IOException ioe) {
            ioe.printStackTrace();
            throw new StoreException("[createFolder] Can't init folder: "
                    + path);
        }
    }

    /**
     * This method removes a folder from the record store. 
     * 
     * @param path
     *             The full pathname in the Store
     * @return <code>true</code> if the folder has been actually deleted,
     *         <code>false</code> if the folder did not exist.
     * @throws StoreException
     *             if an error occurs on the store 
     */
    public boolean removeFolder(String path) throws StoreException {
        path = checkPath(path);
        try {
            // Close the open path, in case was the one to remove
            objs.close();
            RecordStore.deleteRecordStore(path);
        } catch (RecordStoreNotFoundException rsnfe) {
            //Log.debug("removeFolder: folder "+path+" not found.");
            return false;
        } catch (RecordStoreException rse) {
            rse.printStackTrace();
            throw new StoreException("[removeFolder] Can't remove folder: "
                    + path);
        }
        return true;
    }


    /**
     * This method returns the folder whose name is exactly path.
     * 
     * @param path
     *            the path of the folder
     * @return a Folder, or <code>null</code> if not found
     * 
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    public Folder getFolder(String path) throws StoreException {

        path = checkPath(path);
        try {
            objs.open(path);
        } catch (Exception e) {
            e.printStackTrace();
            throw new StoreException("[getFolder] Can't open folder: " + path);
        }
        return new Folder(path, this);
    }


    /**
     * This method returns the list of the folders whose path matches
     * <code>subst</code>.
     * 
     * @param subst
     *            the part of name to search for
     * @return A list of Folder objects whose path matches the provided
     *         substring, or <code>null</code> if there are no matches
     */
    public Folder[] findFolders(String subst) {
        String[] folders = RecordStore.listRecordStores();
        Vector ret = new Vector();

        for (int i = 0; i < folders.length; i++) {
            if (folders[i].startsWith("/") && folders[i].indexOf(subst) != -1)
                ret.addElement(new Folder(folders[i], this));
        }

        return toFolderArray(ret);
    }


    /**
     * This method returns the headers of all the messages in the folder
     * <code>path</code>.
     * 
     * @param path
     *            the path of the folder
     * @return An array of <code>Message</code> with all the headers
     *         set but without content
     * 
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    public Message[] getMsgHeaders(String path) throws StoreException {
        try {
            objs.open(path);

            int count = objs.size();
            Message[] ret = new Message[count];
            if(count ==0) {
                return ret;
            }

            // Load the message headers.
            Serializable s = objs.getFirstObject(new LightMessage());
            ret[0] = (LightMessage) s;
            for (int i = 1; i<count; i++) {
                ret[i] = (LightMessage)objs.getNextObject(new LightMessage());
            }
            return ret;
        } catch (RecordStoreNotFoundException rsnfe) {
            rsnfe.printStackTrace();
            throw new StoreException("[getMsgHeaders] Can't open folder: "
                    + path);
        } catch (RecordStoreException rse) {
            rse.printStackTrace();
            throw new StoreException("[getMsgHeaders] Error accessing folder: "
                    + path);
        } catch (IOException ioe) {
            ioe.printStackTrace();
            throw new StoreException(
                    "[getMsgHeaders] Error getting message headers"
                            + ioe.getMessage());
        }
    }


    /**
     * This method returns the messageIDs of all the messages in the folder
     * <code>path</code>.
     * 
     * @param path
     *            the complete path of the folder
     * @return an array of String, or <code>null</code> if the folder is
     *         empty.
     * 
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    public String[] getMessageIDs(String path) throws StoreException {
        try {
            objs.open(path);
            return objs.names();
        } catch (Exception e) {
            e.printStackTrace();
            throw new StoreException("[getMessageIDs] Can't open folder: "
                    + path);
        }
    }


    // ---------------------------------------------------------- Package
    // methods

    /**
     * This method reads a message from the Store, using the message Id.
     * 
     * @param path
     *            The complete path of the <code>Folder</code> containing the
     *            <code>Message</code>.
     * @param messageId
     *            A string representing the unique message ID
     * @return The <code>Message</code> corresponding to the passed path and
     *         message ID from this <code>Store</code>
     * 
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    public Message readMessage(String path, String messageId)
            throws StoreException {
        try {
            objs.open(path);

            Message m = new Message();
            objs.retrieve(messageId, m);
            return m;
        } catch (RecordStoreNotFoundException rsnfe) {
            rsnfe.printStackTrace();
            throw new StoreException("[readMessage] Can't open folder: " + path);
        } catch (RecordStoreException rse) {
            rse.printStackTrace();
            throw new StoreException("[readMessage] Error accessing folder: "
                    + path);
        } catch (IOException ioe) {
            ioe.printStackTrace();
            throw new StoreException("[readMessage] Error reading message "
                    + messageId);
        }
    }


    /**
     * This method reads the complete message from the Store.
     * 
     * TODO: implement the buffered behaviour for Message, so that a light
     * object is returned by the previous method and the complete one from this,
     * but paying attention to the message state before saving it.
     */
    public Message readMessage(String path, Message msg) throws StoreException {
        // Update the given message
        // TODO: get the content only from the store, and keep the old headers?
        msg = readMessage(path, msg.getMessageId());
        return msg;
    }


    /**
     * This method saves a message in the Store. If the storage left is
     * running low, the older message is removed make room for the new ones.
     * 
     * @param path
     *            The complete path of the <code>Folder</code> containing the
     *            <code>Message</code>.
     * @param msg
     *            The <code>Message</code> to save.
     * 
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    public void saveMessage(String path, Message msg) throws StoreException {
        try {
            objs.open(path);
            // Remove older messages if the storage left is under the
            // threshold.
            if (objs.getAvaliableStorage() < lo_watermark) {
                objs.makeRoom(hi_watermark);
            }
            
            objs.store(msg.getMessageId(), msg);
        } catch (RecordStoreNotFoundException rse) {
            rse.printStackTrace();
            throw new StoreException("[saveMessage] Can't open folder: " + path);
        } catch (RecordStoreException rse) {
            rse.printStackTrace();
            throw new StoreException("[saveMessage] Error accessing folder: "
                    + path);
        } catch (IOException ioe) {
            ioe.printStackTrace();
            throw new StoreException("[saveMessage] Error saving message "
                    + msg.getMessageId());
        }
    }

    /**
     * This method removes a Message from the Store, using message ID as index.
     * 
     * @param path
     *            The complete path of the <code>Folder</code> containing the
     *            <code>Message</code>.
     * @param messageId
     *            A string representing the unique message ID
     * 
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    public void removeMessage(String path, String messageId)
            throws StoreException {
        
        try {
            objs.open(path);
            objs.remove(messageId);
        } catch (RecordStoreException rse) {
            rse.printStackTrace();
            throw new StoreException("[removeMessage] Can't open folder: "
                    + path);
        } catch (IOException ioe) {
            ioe.printStackTrace();
            throw new StoreException("[removeMessage] Error removing message "
                    + messageId);
        }
    }


    /**
     * This method returns the message count in this Store.
     * 
     * @param path
     *            The complete path of the <code>Folder</code> containing the
     *            <code>Message</code>.
     * 
     * @throws StoreException
     *             If an error occurs accessing the Store
     */
    public int countMessages(String path) throws StoreException {
        try {
            objs.open(path);
            return objs.size();
        } catch (Exception e) {
            e.printStackTrace();
            throw new StoreException("[countMessages] Can't open folder: "
                    + path);
        }

    }


    // ---------------------------------------------------------- Private
    // methods

    // Remove the trailing blanks and '/' from path. Add the leading '/'
    // if needed.
    private String checkPath(String path) {
        String ret = path.trim();
        
        if (!ret.startsWith("/")) {
            ret = "/" + ret;
        }
        
        //the normalized path has a "/" at beginning and no "/" at the end
        if (ret.endsWith("/")) {
            return ret.substring(0, ret.length() - 1);
        }
        return ret;
    }


    // Convert a Vector of folders into an array
    private Folder[] toFolderArray(Vector v) {
        int size = v.size();

        if (size == 0) {
            return null;
        }

        Folder[] ret = new Folder[size];

        v.copyInto(ret);
        /*
        for (int i = 0; i < size; i++) {
            ret[i] = (Folder)v.elementAt(i);
        }
         **/
        return ret;
    }

}
